#ifndef DLOADCASE_H
#define DLOADCASE_H

#include "iotools.h"
#include "genfile.h"
#include "alias.h"
#include "seismtool.h"

class dloadn;
class dloadel;
class dloadpd;
class loadcase;

/**
   class dloadcase defines load cases in dynamic problems
   
   JK, 27.9.2004
*/

class dloadcase
{
 public:
  dloadcase (void);
  ~dloadcase (void);
  void read (XFILE *in);
  void print (FILE *out);
  void assemble (long lcid,double *rhs,double *flv,long n, double t);
  void assemble (double *rhs, double *flv,double *lhs);
  void compute_reactions (long lcid);
  void seisminit (double *seism);
  void realloc_pid_array();
  void compute_max_npd();
  double get_pd(double time, long dof);
  double get_macrostre(double time, long dof);
  double get_macrostra(double time, long dof);
  void tempercontrib (long lcid,double *rhs,long n,double t);
  
  ///  type of dynamic load
  dynload tdl;

  ///  the number of subload cases
  long nslc;  
  ///  the number of time dependent loads at nodes
  long nln;
  ///  the number of time dependent loads defined on elements 
  long nle;
  ///  the number of time dependent prescribed displacements
  long npd;
  /// the maximum nuber of prescribed displacements at all subloadcases (used in gradual construction problems)
  long max_npd;
  /// the number of prescribed initial displacements at initial conditions
  long npid;
  /// the number of prescribed initial displacements by rotations
  long nrpid;
  /// the number of all prescribed initial displacements: tnpid = npid+nrpid, it is dimension of array pid
  long tnpid;
  
  /// array of time dependent loads defined at nodes
  dloadn  *lon;
  /// array of time dependent loads defined on elements
  dloadel *loe;
  /// array of time dependent prescribed displacements defined at nodes
  dloadpd *pd;
  /// array of prescribed initial displacements at nodes for gradual contruction problem
  double  *pid; 
  
  /// array of subload cases
  loadcase *slc;
  /// array of time functions
  gfunct *gf;
  ///  seismic tool
  seismtool stool;

};

#endif
