#include <errno.h>
#include "globmat.h"
#include "genfile.h"
#include "mathem.h"
#include "global.h"
#include "elemhead.h"
#include "loadcase.h"
#include "dloadcase.h"
#include "dloadpd.h"
#include "node.h"
#include "element.h"
#include "intpoints.h"
#include "problem.h"
#include "mechmat.h"
#include "creepb.h"
#include "creepbs.h"
#include "mechcrsec.h"
#include "elemswitch.h"
#include "vecttens.h"


/**
  Function assembles stiffness %matrix.
  
  Parameters: 
  @param lcid - load case id
   
  @return The function does not return anything.

  Created by JK,
*/
void stiffness_matrix (long lcid)
{
  long i,j,nl,ndofe,ndofemn,ndofn,nmult,nid;
  ivector cn,cna;
  matrix lm,lmt;
  matrix alm,bd,db,dd;
  time_t bt,et;
  int err;
  
  bt = time (NULL);

  if (Smat==NULL)
    Smat = new gmatrix;

  Smat->setval (Mp->ssle);
  Smat->initiate (Gtm,Ndofm,Mp->tstorsm,Mespr,Out);
  
  //  the following condition is used only for problem type which
  //  deals with load balancing method
  //  the code stops here because it needs to compute number of matrix
  //  entries in the system matrix
  if (Mp->lbtype==1)
    abort ();
  
  
  
  //FILE *ma;
  //ma=fopen ("matice.txt","w");

  
  for (i=0;i<Mt->ne;i++){
    if (Gtm->leso[i]==1){
      //  only active elements are assembled
      //  this is used especially in problems with changing structures
      
      //  the number of DOFs on element
      //  this number is equal to the number of DOFs without hanging nodes
      ndofe = Mt->give_ndofe (i);
      
      reallocm (ndofe,ndofe,lm);
      //  computation of stiffness matrix of one element
      stiffmat (lcid,i,lm);

      err =check_math_errel(i);

      //  the number of DOFs on element
      //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
      ndofemn = Gtm->give_ndofe (i);
      reallocv (ndofemn,cn);
      Mt->give_code_numbers (i,cn.a);
      
      if (ndofe != ndofemn){
	if (Mp->homog!=3){
	  //  this case means hanging nodes
	  
	  //  the element contains hanging node
	  //  the stiffness matrix has to be transformed
	  reallocm (ndofe,ndofemn,alm);
	  mxm (lm,*Mt->elements[i].tmat,alm);
	  reallocm (ndofemn,ndofemn,lm);
	  mtxm (*Mt->elements[i].tmat,alm,lm);
	}else{
	  //  this case means homogenization
	  
	  long ii,jj,kk;
	  reallocm (ndofe,ndofe,alm);
	  copym (lm,alm);
	  reallocm (ndofemn,ndofemn,lm);	  
	  for (ii=0;ii<ndofe;ii++){
	    for (jj=0;jj<ndofe;jj++){
	      lm(ii, jj)=alm(ii, jj);
	    }
	  }
	  reallocm (ndofe,ndofemn-ndofe,bd);
	  Ltet->bd_matrix (i,0,0,bd);//co to je za klauni? jenom lintet?
	  for (ii=0;ii<ndofe;ii++){
	    kk=ndofe;
	    for (jj=0;jj<ndofemn-ndofe;jj++){
	      lm(ii, kk)=bd(ii, jj);
	      kk++;
	    }
	  }
	  kk=ndofe;
	  for (ii=0;ii<ndofemn-ndofe;ii++){
	    for (jj=0;jj<ndofe;jj++){
	      lm(kk, jj)=bd(jj, ii);
	    }
	    kk++;
	  }
	  reallocm (ndofemn-ndofe,ndofemn-ndofe,dd);
	  Ltet->dd_matrix (i,0,0,dd);//co to je za klauni? jenom lintet?
	  for (ii=0;ii<ndofemn-ndofe;ii++){
	    for (jj=0;jj<ndofemn-ndofe;jj++){
	      lm(ii+ndofe, jj+ndofe)=dd(ii, jj);
	    }
	  }
	  
	}
      } 
      //  localization of matrix of one element to matrix of problem
      Smat->localize (lm,cn,i);
    }
  }
  
  //Smat->printmat (ma);
  //fclose (ma);
  
  /*
    fclose (ma);
    
    ma=fopen ("sit.txt","w");
    fprintf (ma,"%ld\n",Gtm->nn);
    for (i=0;i<Gtm->nn;i++){
    fprintf (ma,"%ld  %lf %lf %lf  %ld %ld %ld\n",i+1,Gtm->gnodes[i].x,Gtm->gnodes[i].y,Gtm->gnodes[i].z,Gtm->gnodes[i].cn[0],Gtm->gnodes[i].cn[1],Gtm->gnodes[i].cn[2]);
    }
    fprintf (ma,"\n\n%ld\n",Gtm->ne);
    for (i=0;i<Gtm->ne;i++){
    fprintf (ma,"%ld    %ld %ld %ld %ld\n",i+1,Gtm->gelements[i].nodes[0],Gtm->gelements[i].nodes[1],Gtm->gelements[i].nodes[2],Gtm->gelements[i].nodes[3]);
    //fprintf (ma,"%ld    %ld %ld %ld %ld %ld %ld %ld %ld\n",
    //i+1,Gtm->gelements[i].nodes[0],Gtm->gelements[i].nodes[1],Gtm->gelements[i].nodes[2],
    //Gtm->gelements[i].nodes[3],Gtm->gelements[i].nodes[4],Gtm->gelements[i].nodes[5],
    //Gtm->gelements[i].nodes[6],Gtm->gelements[i].nodes[7]);
    }
    fclose (ma);
  */
  
  
  
  /*
    DSS::SparseMatrixF sm((unsigned long)Ndofm,Smat->scr->a,(unsigned long*)Smat->scr->ci,(unsigned long*)Smat->scr->adr);
    FILE *out;
    out = fopen ("matice.sm","wb");
    sm.SaveMatrix(out);
  */
  
  if (Mp->tprob==layered_linear_statics){
    for (i=0;i<Mt->nln;i++){
      nl=Gtm->lgnodes[i].nl;
      ndofn=Gtm->lgnodes[i].ndofn;
      nmult=Gtm->lgnodes[i].nmult;
      //allocm (nmult,ndofn*2,lm);
      reallocm (ndofn*2,nmult,lm);
      reallocm (nmult,ndofn*2,lmt);
      //cn = new long [ndofn*2];
      reallocv (ndofn*2,cn);
      //cna = new long [nmult];
      reallocv (nmult,cna);
    
      for (j=1;j<nl;j++){
	constr_matrix (i,j,lm);
	
	nid=Gtm->lgnodes[i].nodes[j-1];
	Mt->give_node_code_numbers (nid,cn.a);
	nid=Gtm->lgnodes[i].nodes[j];
	Mt->give_node_code_numbers (nid,cn.a+ndofn);
	Mt->give_mult_code_numbers (i,j,cna.a);
	
	tranm (lm,lmt);
	
	Smat->glocalize (lmt,cna,cn);
	Smat->glocalize (lm,cn,cna);
	
	
      }
    }
  }
  
  Smat->prepmat (0.0,Mespr);
  
  //Smat->printdiag(Out);
  //Smat->printmat(Out);
  
  et = time (NULL);
//  fprintf (stdout,"\n stiffness matrix assembling time    %ld",et-bt);
  fflush(stdout);
  
}



/**
  Function assembles mass %matrix.
   
  Parameters:
  @param lcid - load case id
   
  @return The function does not return anything.

  Created by JK,
*/
void mass_matrix (long lcid)
{
  long i,ndofe,ndofemn;
  ivector cn;
  matrix lm;
  matrix alm;
  
  if (Mmat==NULL)  Mmat = new gmatrix;
  //Mmat->ts=Mp->tstormm;
  Mmat->setval (Mp->ssle);
  Mmat->initiate (Gtm,Ndofm,Mp->tstormm,Mespr,Out);
  //Mmat->setval (Mp->ssle.tlinsol,Mp->ssle.prec.pt,Mp->ssle.ni,Mp->ssle.err);
  
  
  for (i=0;i<Mt->ne;i++){
    if (Gtm->leso[i]==1){
      //  number of DOFs on the i-th element
      ndofe = Mt->give_ndofe (i);
      //  allocation of element mass matrix
      reallocm (ndofe,ndofe,lm);
      //  computation of mass matrix of one element
      massmat (lcid,i,lm);
      


      //  the number of DOFs on element
      //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
      ndofemn = Gtm->give_ndofe (i);
      reallocv (ndofemn,cn);
      Mt->give_code_numbers (i,cn.a);
      
      if (ndofe != ndofemn){
	//  the element contains hanging node
	//  the stiffness matrix has to be transformed
	reallocm (ndofe,ndofemn,alm);
	mxm (lm,*Mt->elements[i].tmat,alm);
	reallocm (ndofemn,ndofemn,lm);
	mtxm (*Mt->elements[i].tmat,alm,lm);
      }
      
      //  localization of matrix of element to matrix of problem
      Mmat->localize (lm,cn,i);
    }
  }
  
  /*
  for (i=0;i<Mt->nn;i++){
    cr = Mt->nodes[i].crst;
    if (cr==nocrosssection){
      continue;
    }
    else{
      ndofn = Mt->give_ndofn (i);
      for (j=0;j<ndofn;j++){
	k = Mt->give_dof (i,j);
	if (k>0){
	  idcs = Mt->nodes[i].idcs;
	  m=Mc->give_weight (cr,idcs);
	  Mmat->add_entry (m,k-1,k-1);
	}
      }
    }
  }
  */

  Mmat->prepmat (0.0,Mespr);
}



/**
  Function assembles initial stiffness %matrix.

  It is also called geometric stiffness %matrix
  (in R.W. Clough and J. Penzien: Dynamics of structures)

  Parameters:   
  @param lcid - load case id
   
  @return The function does not return anything.

  Created by JK,
*/
void initial_stiffness_matrix (long lcid)
{
  if (Ismat==NULL)  Ismat = new gmatrix;
  //Ismat->ts=Mp->tstorsm;
  Ismat->setval (Mp->ssle);
  Ismat->initiate (Gtm,Ndofm,Mp->tstorsm,Mespr,Out);
  //Ismat->setval (Mp->ssle.tlinsol,Mp->ssle.prec.pt,Mp->ssle.ni,Mp->ssle.err);
  
  
  long i,ndofe,ndofemn;
  ivector cn;
  matrix lm;
  matrix alm;
  
  for (i=0;i<Mt->ne;i++){
    if (Gtm->leso[i]==1){

      ndofe = Mt->give_ndofe (i);
      
      reallocm (ndofe,ndofe,lm);
      
      //  computation of initial stiffness matrix of one element
      initstiffmat (lcid,i,lm);
      
      //  the number of DOFs on element
      //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
      ndofemn = Gtm->give_ndofe (i);
      reallocv (ndofemn,cn);
      //cn = new long [ndofe];
      Mt->give_code_numbers (i,cn.a);
      
      if (ndofe != ndofemn){
	//  the element contains hanging node
	//  the stiffness matrix has to be transformed
	reallocm (ndofe,ndofemn,alm);
	mxm (lm,*Mt->elements[i].tmat,alm);
	reallocm (ndofemn,ndofemn,lm);
	mtxm (*Mt->elements[i].tmat,alm,lm);
      }
      
      //  localization of matrix of one element to matrix of problem
      Ismat->localize (lm,cn,i);
    }
  }
  
  Ismat->prepmat (0.0,Mespr);
}



/**
  The function assembles damping %matrix.
   
  @return The function does not return anything.

  Created by JK, 10.8.2005
*/
void damping_matrix ()
{
  if (Dmat==NULL)  Dmat = new gmatrix;
  //Dmat->ts=Mp->tstorsm;
  Dmat->setval (Mp->ssle);
  Dmat->initiate (Gtm,Ndofm,Mp->tstorsm,Mespr,Out);
  //Dmat->setval (Mp->ssle.tlinsol,Mp->ssle.prec.pt,Mp->ssle.ni,Mp->ssle.err);
  
  switch (Mp->damp){
  case nodamp:{
    break;
  }
  case massdamp:{
    Dmat->addgm (Mp->dmass,*Mmat);
    break;
  }
  case stiffdamp:{
    Dmat->addgm (Mp->dstiff,*Smat);
    break;
  }
  case rayleighdamp:{
    Dmat->addgm (Mp->dmass,*Mmat);
    Dmat->addgm (Mp->dstiff,*Smat);
    break;
  }
  default:{
    print_err("unknown type of damping is required",__FILE__,__LINE__,__func__);
  }
  }
  
  Dmat->prepmat (0.0,Mespr);
}



/**
  The function computes %vector of internal forces.
   
  @param lcid - load case id
  @param intfor - %vetcor of internal forces (output)
   
  @return The function returns %vector of internal forces
  in the parameter intfor.

  Created by JK, 28.7.2001
*/
void internal_forces (long lcid,double *intfor)
{
  if (Mp->matmodel==local){
    loc_internal_forces (lcid,intfor);
  }
  if (Mp->matmodel==nonlocal){
    nonloc_internal_forces (lcid,intfor);
  }
  
  //  contribution to internal forces by Lagrange multipliers
  //  it is used in problems with floating subdomains solved by the FETI method
  //  JK, 22.6.2006
  if (Mp->tprob == lin_floating_subdomain || Mp->tprob == nonlin_floating_subdomain){
    //lagrmultcontr_intforces (lcid,intfor);
  }

  
  //  indicator of strain computation
  //  it means, strains at integration points have been computed
  Mp->strainstate=1;
  //  indicator of stress computation
  //  it means, stresses at integration points have been computed
  Mp->stressstate=1;
  //  indicator of computation of other array
  //  it means, stresses at integration points have been computed
  Mp->otherstate=1;
}



/**
  The function computes %vector of internal forces,
  local material models are used.

  @param lcid - load case id
  @param intfor - %vector of internal forces (output)
   
  @return The function returns %vector of internal forces computed localy
  in the parameter intfor.

  Created by JK, TKo,  28.7.2001
*/
void loc_internal_forces (long lcid,double *intfor)
{
#ifndef INC_OPENMP
  long i,j,ndofe,ndofemn, ncomp=0L;
  ivector cn;
  int err;
  vector ifor, aifor;
  vector r, mstr;
  matrix bd, dd;
#endif

  //  function computes strains at integration points
  if (Mp->strcomp)
    compute_ipstrains (lcid);
    
  //  number of elements
  long ne=Mt->ne;
  
  //  cleaning of the array intfor 
  nullv (intfor,Ndofm);
#ifdef INC_OPENMP
#pragma omp parallel num_threads(4)
 {
  long i,j,ndofe,ndofemn, ncomp=0L;
  ivector cn;
  int err;
  vector ifor, aifor;
  vector r, mstr;
  matrix bd, dd;

  #pragma omp for nowait
#endif
  
  for (i=0;i<ne;i++){
    if (Gtm->leso[i]==1){
      
      //  the number of DOFs on element
      //  this number is equal to the number of DOFs without hanging nodes
      ndofe=Mt->give_ndofe (i);
      reallocv (RSTCKVEC(ndofe,ifor));
      
      elem_internal_forces (i,lcid,ifor);

      err = check_math_errel(i);
      
      if (err)
      {
        abort();
      }
      
      //  the number of DOFs on element
      //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
      ndofemn = Gtm->give_ndofe (i);
      reallocv (RSTCKIVEC(ndofemn,cn));

      Mt->give_code_numbers (i,cn.a);

      if (ndofe != ndofemn)
      {
	if (Mp->homog!=3)
        {
          //  this case means hanging nodes

          //  the element contains hanging node
          //  the stiffness matrix has to be transformed
          reallocv (RSTCKVEC(ndofemn,aifor));
          mtxv (*Mt->elements[i].tmat,ifor,aifor);
         #ifdef INC_OPENMP
          #pragma omp critical(locglob)
          {
         #endif
          locglob (intfor,aifor.a,cn.a,ndofemn);
         #ifdef INC_OPENMP
          }
         #endif
	}
        else
        {
	  //  this case means homogenization
          if (ncomp == 0L)
	    ncomp = ndofemn-ndofe;

          // localize contribution from standard element stresses due 
          // to macro-strains and fluctuations 
          // (macro-strains are added in compute_ipstrains function)
         #ifdef INC_OPENMP
          #pragma omp critical(locglob)
          {
         #endif
          locglob (intfor,ifor.a,cn.a,ndofe);
         #ifdef INC_OPENMP
          }
         #endif
          // contribution to macro-stresses due to macro-strains and fluctuations
          // (macro-strains are added in compute_ipstrains function)
          reallocv (RSTCKVEC(ncomp, ifor));
          elem_integration_quant(i, locstress, lcid, ifor);
         #ifdef INC_OPENMP
          #pragma omp critical(locglob)
          {
         #endif
          for (j=0; j<ncomp; j++)
            intfor[Ndofm-ncomp+j] += ifor[j];             
         #ifdef INC_OPENMP
          }
         #endif
	}
      }
      else
      {
       #ifdef INC_OPENMP
        #pragma omp critical(locglob)
        {
       #endif
	//  the element does not contain hanging node or homogenization
	locglob (intfor,ifor.a,cn.a,ndofe);
       #ifdef INC_OPENMP
        }
       #endif
      }
    }
  }
#ifdef INC_OPENMP
 }
#endif
}



/**
  The function computes %vector of internal forces,
  nonlocal material models are used.

  @param lcid - load case id
  @param intfor - %vector of internal forces (output)

  @return The function returns %vector of internal forces computed nonlocaly
  in the parameter intfor.

  Created by JK,TKo, 28.7.2001
*/
void nonloc_internal_forces (long lcid,double *intfor)
{
  long i,ne,ndofe,ndofemn;
  ivector cn;
  vector ifor;
  vector aifor;

  //  function computes strains at integration points
  if (Mp->strcomp)
    compute_ipstrains (lcid);

  ne=Mt->ne;
  nullv (intfor,Ndofm);
  Mp->nonlocphase=1;

  for (i = 0; i < ne; i++)
  {
    if (Gtm->leso[i]==1)
     elem_local_values (i, lcid);
  }
  
  for (i = 0; i < Mm->tnip; i++)
    Mm->nonlocaverage(i);
  Mp->nonlocphase=2;
  
  for (i = 0; i < ne; i++)
  {
    if (Gtm->leso[i]==1)
    {
      //  the number of DOFs on element
      //  this number is equal to the number of DOFs without hanging nodes
      ndofe=Mt->give_ndofe (i);
      reallocv (ndofe,ifor);

      elem_nonloc_internal_forces(i, lcid, ifor);

      check_math_errel(i);

      //  the number of DOFs on element
      //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
      ndofemn = Gtm->give_ndofe (i);
      reallocv (ndofemn,cn);

      Mt->give_code_numbers (i,cn.a);
      if (ndofe != ndofemn){
	//  the element contains hanging node
	//  the stiffness matrix has to be transformed
	reallocv (ndofemn,aifor);
	mtxv (*Mt->elements[i].tmat,ifor,aifor);
	locglob (intfor,aifor.a,cn.a,ndofemn);
      }else{
	//  the element does not contain hanging node
	locglob (intfor,ifor.a,cn.a,ndofe);
      }
    }
  }
}



/**
  The function computes %vector of increments of internal forces.
   
  @param lcid - load case id
  @param intfor - %vector of increments of internal forces (output)
   
  @return The function returns %vector of increments of internal forces 
  in the parameter intfor.

  Created by JK, TKo, 29.4.2008
*/
void incr_internal_forces (long lcid,double *intfor)
{
  long i,ne,ndofe,ndofemn;
  ivector cn;
  vector ifor;
  vector aifor;
  
  //  function computes strains at integration points
  compute_ipstrains (lcid);
    
  //  number of elements
  ne=Mt->ne;
  
  //  cleaning of the array intfor
  nullv (intfor,Ndofm);
  
  for (i=0;i<ne;i++){
    if (Gtm->leso[i]==1){
      
      //  the number of DOFs on element
      //  this number is equal to the number of DOFs without hanging nodes
      ndofe=Mt->give_ndofe (i);
      reallocv (ndofe,ifor);
      
      elem_incr_internal_forces (i,lcid,ifor);

      check_math_errel(i);
      
      //  the number of DOFs on element
      //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
      ndofemn = Gtm->give_ndofe (i);
      reallocv (ndofemn,cn);
      
      Mt->give_code_numbers (i,cn.a);

      if (ndofe != ndofemn){
	//  the element contains hanging node
	//  the stiffness matrix has to be transformed
	reallocv (ndofemn,aifor);
	mtxv (*Mt->elements[i].tmat,ifor,aifor);
	locglob (intfor,aifor.a,cn.a,ndofemn);
      }else{
	//  the element does not contain hanging node
	locglob (intfor,ifor.a,cn.a,ndofe);
      }
    }
  }
}



/**
  Function adds contributions to internal forces caused by
  Lagrange multipliers. Function is used in problems with floating subdomains
  solved by the FETI method.
   
  @param lcid - load case id
  @param intfor - array of internal forces (input/output)
   
  @return The function adds contributions to the parameter intfor.

  Created by JK, 22.6.2006
*/
/*
void lagrmultcontr_intforces (long lcid,double *intfor)
{
  long i;
  double *av,*av1,*av2,*cv;
  
  av = new double [Ndofm];
  av1 = new double [Ndofm];
  av2 = new double [Ndofm];
  cv = new double [Fsd->nlm];
  
  //  E^T lamba
  nullv (av,Ndofm);
  nullv (av1,Ndofm);
  nullv (av2,Ndofm);
  //Gtm->flsub.coarse_local (Fsd->tw,av);
  Gtm->flsub.coarse_local (Fsd->muo,av1);
  Gtm->flsub.coarse_local (Fsd->ddmu,av2);
  
  for (i=0;i<Ndofm;i++){
    av[i]=av1[i]+av2[i];
  }
*/
  //Gtm->flsub.local_coarse (cv,Lsrs->lhs);
  
  /*
  long i;
  for (i=0;i<Ndofm;i++){
    fprintf (Out,"\n lagr mult %le",av[i]);
  }
  for (i=0;i<Fsd->nlm;i++){
    fprintf (Out,"\n coarse vector %le",cv[i]);
    

    if (cv[i]>10.0){
      if (av[i]<0.0){
	av[i]=-10.0;
	Fsd->tw[i]=-10.0;
      }
      if (av[i]>0.0){
	av[i]=10.0;
	Fsd->tw[i]=10.0;
      }
      Fsd->compli[i]=2.0;
    }
    
  }
  */
/*
  addv (intfor,av,Ndofm);
  
  delete [] av;
  delete [] cv;
}
*/


/**
   The function computes %vector of nodal forces
   caused by eigenstrains. Eigenstrains may contain 
   either eigenstrains or temperature strains.
   
   @param lcid - load case id
   @param nodfor - %vector of nodal forces (output)
   @param time - actual time
   
   @return The fucntion returns nodal forces in the parameter nodfor.
   
   Created by JK, 28.7.2001
*/
void nodal_eigstrain_forces (long lcid,double *nodfor,double time)
{
  long i,ne,ndofe,ndofemn;
  ivector cn;
  vector nfor;
  vector anfor;
  
  //  array nodfor must not be cleaned
  //  it contains contributions from previous calculations
  //  this function should only contribute to this array
  
  if (Mp->eigstrains==1 || Mp->eigstrains==2) 
    Mb->eigstrain_computation (time);
  
  if (Mp->eigstrains==4 || Mp->eigstrains==5) 
  {
    Mb->eigstrain_computation (time);
    Mp->eigstrcomp = 0; 
  }

  //  the number of elements
  ne=Mt->ne;
  
  for (i=0;i<ne;i++){
    if (Gtm->leso[i]==1){
      
      //  the number of DOFs on element
      //  this number is equal to the number of DOFs without hanging nodes
      ndofe=Mt->give_ndofe (i);
      reallocv (ndofe,nfor);

      //  computes contribution from one element
      elem_eigstrain_forces (lcid,i,nfor);

      // change sign of contribution nfor from element because loacalization 
      // procedure adds it to the global vector and the resulting global vector is 
      // added to the global right hand side vector
      // K_e r_e = f_e - \int\limits_\Omega_e B^T D_e \eps_0
      // K_e r_e = f_e - \int\limits_\Omega_e B^T \sigma_0
      // K_e r_e = f_e - nfor
      chsgnv(nfor);

      check_math_errel(i);

      //  the number of DOFs on element
      //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
      ndofemn = Gtm->give_ndofe (i);
      reallocv (ndofemn,cn);
      
      Mt->give_code_numbers (i,cn.a);

      if (ndofe != ndofemn){
	//  the element contains hanging node
	//  the vector has to be transformed
	reallocv (ndofemn,anfor);
	mtxv (*Mt->elements[i].tmat,nfor,anfor);
	locglob (nodfor,anfor.a,cn.a,ndofemn);
      }else{
	//  the element does not contain hanging node
	locglob (nodfor,nfor.a,cn.a,ndofe);
      }
    }
  }
}



/**
   The function computes %vector of nodal forces
   caused by pore pressures. The function is employed only for
   coupled problems where the pore pressures are defined in TRFEL 
   out of the MEFEL control.
   
   
   @param lcid - load case id
   @param nodfor - %vector of nodal forces (output)
   @param time - actual time
   
   @return The fucntion returns nodal forces in the parameter nodfor.
   
   Created by JK, 28.7.2001
*/
void nodal_pore_press_forces (long lcid,double *nodfor,double time)
{
  long i,ne,ndofe,ndofemn, besc;
  ivector cn;
  vector nfor;
  vector anfor;
  double u;        // pore pressure
  vector sigt(6);
  //matrix sigt(3,3);
  vector sig;
  
  //  array nodfor must not be cleaned
  //  it contains contributions from previous calculations
  //  this function should only contribute to this array
  
  for(i=0; i<Mm->tnip; i++)
  {
    u = Mm->givenonmechq(pore_pressure, i);
    reallocv(Mm->ip[i].ncompstr, sig);
    sigt[0] = sigt[1] = sigt[2] = u;
    give_red_vector(sigt,sig,Mm->ip[i].ssst);
    Mm->storeeigstress(i, sig);
  }

  // eigenstresses have been already defined above and they must not be rewritten
  besc = Mp->eigstrcomp; // backup of eigenstress compuation flag
  Mp->eigstrcomp = 0; 

  //  the number of elements
  ne=Mt->ne;
  
  for (i=0;i<ne;i++){
    if (Gtm->leso[i]==1){
      
      //  the number of DOFs on element
      //  this number is equal to the number of DOFs without hanging nodes
      ndofe=Mt->give_ndofe (i);
      reallocv (ndofe,nfor);
      
      //  computes contribution from one element
      elem_eigstrain_forces (lcid,i,nfor);

      // change sign of contribution nfor from element because loacalization 
      // procedure adds it to the global vector and the resulting global vector is 
      // added to the global right hand side vector
      // K_e r_e = f_e - \int\limits_\Omega_e B^T u where u  is the vector of pore pressures
      // K_e r_e = f_e - nfor
      chsgnv(nfor);

      check_math_errel(i);

      //  the number of DOFs on element
      //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
      ndofemn = Gtm->give_ndofe (i);
      reallocv (ndofemn,cn);
      
      Mt->give_code_numbers (i,cn.a);

      if (ndofe != ndofemn){
	//  the element contains hanging node
	//  the vector has to be transformed
	reallocv (ndofemn,anfor);
	mtxv (*Mt->elements[i].tmat,nfor,anfor);
	locglob (nodfor,anfor.a,cn.a,ndofemn);
      }else{
	//  the element does not contain hanging node
	locglob (nodfor,nfor.a,cn.a,ndofe);
      }
    }
  }

  // restore original status of eigenstress computation flag
  Mp->eigstrcomp = besc; 
}



/**
  The function extracts displacements on one element.
   
  @param lcid - number of load case
  @param eid - element id
  @param r - allocated array for displacement, it is output parameter
   
  @return The function returns nodal displacements at the given element in the array r.

  Created by JK, 9.7.2001
*/
void eldispl (long lcid,long eid,double *r)
{
  long i,j,ii,ndofe,ndofemn,nne,ndofn,acn;
  ivector cn,nod,ncn;
  vector rr;
  
  //  the number of DOFs on element
  //  this number is equal to the number of DOFs without hanging nodes
  ndofe = Mt->give_ndofe (eid);
  //  the number of DOFs on element
  //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
  ndofemn = Gtm->give_ndofe (eid);
  
  if (ndofe != ndofemn){
    //  there are hanging nodes on element
    reallocv (ndofemn,cn);
    reallocv (ndofemn,rr);
    ndofe=ndofemn;
  }
  else{
    //  there are no hanging nodes on element
    reallocv (ndofe,cn);
    reallocv (ndofe,rr);
  }
  //  code numbers on element
  Mt->give_code_numbers (eid,cn.a);
  
  
  switch (Mp->tprob){
  case forced_dynamics:
  case mech_timedependent_prob:{
    //  lcid must be equal to zero
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      if (ii<0)   rr[i]=Mb->dlc[lcid].get_pd(Mp->time, ii);
      if (ii==0)  rr[i]=0.0;
      if (ii>0)   rr[i]=Lsrs->lhs[lcid*Ndofm+ii-1];
    }
    break;
  }

  case mat_nonlinear_statics:{
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      if (ii<0)
      {   
        rr[i] = 0.0;
        if (Mb->lc[lcid].npd)
          rr[i] += Mb->lc[lcid].pd[0-ii-1] * Mp->lambda;
        if (Mb->lc[lcid+1].npd)
          rr[i]+= Mb->lc[lcid+1].pd[0-ii-1];
      }
      if (ii==0)  rr[i]=0.0;
      if (ii>0)   rr[i]=Lsrs->lhs[lcid*Ndofm+ii-1];
    }
    break;
  }
    
  case growing_mech_structure:{
    //  lcid must be equal to zero
    
    //  the number of nodes on element
    //  if there are hanging nodes, the master nodes are included
    nne = Gtm->give_nne(eid);
    
    reallocv (nne,nod);
    //  nodes on element
    Gtm->give_nodes (eid,nod);
    
    acn=0;
    //  loop over the number of nodes on element
    for (i=0;i<nne;i++){
      
      //  the number of DOFs on node
      ndofn = Gtm->give_ndofn (nod[i]);
      
      reallocv (ndofn,ncn);
      //  code numbers in node
      Gtm->give_node_code_numbers (nod[i],ncn.a);
      
      //  loop over the number of DOFs in node
      for (j=0;j<ndofn;j++){
	ii=ncn[j];
	if (ii<0)   rr[acn+j]=Mb->dlc[lcid].get_pd(Mp->time, ii);
	if (ii==0)  rr[acn+j]=0.0;
	if (ii>0)   rr[acn+j]=Lsrs->lhs[lcid*Ndofm+ii-1];
	
        if (Mt->nodedispl[nod[i]] && (ii < 1)) 
	  // add initial displacements from previous free state of dof 
	  // in the case that the dof has got support or prescribed displacement actually
          rr[acn+j] += Mt->nodedispl[nod[i]][j];
      }//  end of the loop over the number of DOFs in node
      acn += ndofn;
    }//  end of the loop over the number of nodes on element
    
    break;
  }
    
  default:{
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      if (ii<0)   rr[i]=Mb->lc[lcid].pd[0-ii-1];
      if (ii==0)  rr[i]=0.0;
      if (ii>0)   rr[i]=Lsrs->lhs[lcid*Ndofm+ii-1];
    }
  }
  }//  end of the switch (Mp->tprob)
  
  //  the variables ndofe and ndofemn have to be obtained again
  //  because they are possibly rewritten
  //
  //  the number of DOFs on element
  //  this number is equal to the number of DOFs without hanging nodes
  ndofe = Mt->give_ndofe (eid);
  //  the number of DOFs on element
  //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
  ndofemn = Gtm->give_ndofe (eid);

  if (ndofe != ndofemn){
    if (Mp->homog!=3){
      //  this case means hanging nodes
      
      //  there are hanging nodes on element
      mxv (Mt->elements[eid].tmat->a,rr.a,r,ndofe,ndofemn);
    }
    else{
      //  this case means homogenization
      //  the vector rr contains more components than the vector r
      //  only the "classical" components are copied
      copyv (rr.a,r,ndofe);
    }
  }
  else{
    //  there are no hanging nodes on element
    copyv (rr,r);
  }
  
  if (Mp->tprob == growing_mech_structure){
    //  subtraction of initial displacements
    Mt->elements[eid].subtrinitdispl (r, ndofe);
  }
  
  
  destrv (rr);
  destrv (cn);
}



/**
  The function extracts prescribed displacements on one element.
   
  @param lcid - number of load case
  @param eid - element id
  @param r - array of prescribed displacements, it is output parameter
  
  @return The function returns extracted prescribed displacements in the array r.

  Created by JK, 9.7.2001
*/
 void elprdispl (long lcid,long eid,double *r)
{
  long i,ii,ndofe,ndofemn;
  ivector cn,nod,ncn;
  vector rr;
  
  //  the number of DOFs on element
  //  this number is equal to the number of DOFs without hanging nodes
  ndofe = Mt->give_ndofe (eid);
  //  the number of DOFs on element
  //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
  ndofemn = Gtm->give_ndofe (eid);
  
  if (ndofe != ndofemn){
    //  there are hanging nodes on element
    reallocv (ndofemn,cn);
    reallocv (ndofemn,rr);
    ndofe=ndofemn;
  }
  else{
    //  there are no hanging nodes on element
    reallocv (ndofe,cn);
    reallocv (ndofe,rr);
  }
  //  code numbers on element
  Mt->give_code_numbers (eid,cn.a);
  
  switch (Mp->tprob){
  case forced_dynamics:
  case mech_timedependent_prob:{
    //  lcid must be equal to zero
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      if (ii<0)   
        rr[i]=Mb->dlc[lcid].get_pd(Mp->time, ii);
    }
    break;
  }

  case mat_nonlinear_statics:{
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      if (ii<0)
      {   
        rr[i]=Mb->lc[lcid].pd[0-ii-1];
        if ((lcid % 2) == 0) // the proportional load case is each 1-st, 3-rd, 5-th, ... => lcid = 0, 2, 4,...
          rr[i]*=Mp->lambda;
      } 
    }
    break;
  }
    
  case growing_mech_structure:{
    //  lcid must be equal to zero
    
    //  loop over the number of DOFs on element (including DOFs in hanging nodes)
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      if (ii<0)   
        rr[i]=Mb->dlc[lcid].get_pd(Mp->time, ii);
    }    
    // initial displacements from previous free state of dof 
    // in the case that the dof has got support or prescribed displacement actually
    // must not be included to the computation of the right hand side because 
    // the displacements of the neighbour free dofs are already included in the 
    // displacement vector (r) as well as in the load vector from the previous time step (fp)
    break;
  }
    
  default:{
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      if (ii<0)   
        rr[i]=Mb->lc[lcid].pd[0-ii-1];
    }
  }
  }//  end of the switch (Mp->tprob)
  
  //  the variables ndofe and ndofemn have to be obtained again
  //  because they are possibly rewritten
  //
  //  the number of DOFs on element
  //  this number is equal to the number of DOFs without hanging nodes
  ndofe = Mt->give_ndofe (eid);
  //  the number of DOFs on element
  //  this number is equal to the number of DOFs with master nodes generated by hanging nodes
  ndofemn = Gtm->give_ndofe (eid);
  
  if (ndofe != ndofemn){
    //  there are hanging nodes on element
    mxv (Mt->elements[eid].tmat->a,rr.a,r,ndofe,ndofemn);
  }
  else{
    //  there are no hanging nodes on element
    copyv (rr,r);
  }
  
  
  if (Mp->tprob == growing_mech_structure){
    //  subtraction of initial displacements
    Mt->elements[eid].subtrinitdispl (r,ndofe);
  }
  
}



/**
  The function extracts displacements on one node with the index nid and stores them in the %vector r.

  @param lcid  - number of load case
  @param r     - allocated array for displacement, it is output parameter
  @param nid   - node number (node id)

  @return The function returns nodal displacements at the given node in the array r.

  Created by JK 9.7.2001
  Modified by Tomas Koudelka
*/
void noddispl (long lcid,double *r, long nid)
{
  long i,j,ndofn,ndofnm,nmn;
  double **rr;
  vector w;
  
  //  the number of DOFs in node
  ndofn = Gtm->give_ndofn(nid);
  
  if (ndofn<0){
    //  this is a hanging node
    
    //  the number of master nodes
    nmn=0-ndofn;
    
    rr = new double* [nmn];
    //  loop over the number of master nodes
    for (i=0;i<nmn;i++){
      //  the number of DOFs on master node
      ndofnm = Gtm->give_ndofn (Gtm->gnodes[nid].mnodes[i]);
      rr[i] = new double [ndofnm];
      select_noddispl (lcid,rr[i],Gtm->gnodes[nid].mnodes[i]);
    }
    
    //  weights
    allocv (nmn,w);
    Gtm->approx_weights (Gtm->gnodes[nid].masentity,nid,w,Out);
    
    //  approximation
    //r = new double [ndofnm];
    for (i=0;i<ndofnm;i++){
      r[i]=0.0;
      for (j=0;j<nmn;j++){
	r[i]+=rr[j][i]*w[j];
      }
    }
    
    for (i=0;i<0-ndofn;i++){
      delete [] rr[i];
    }
    delete [] rr;
    
    ndofn = ndofnm;
  }else{
    //  this is not a hanging node
    
    //r = new double [ndofn];
    
    select_noddispl (lcid,r,nid);
  }//  end of the statement if (ndofn<0){
  
  //return ndofn;
}



/**
   The function selects nodal displacements of the nid-th node and stores them
   into the %vector r.
   
   @param lcid - load case id
   @param r - array of nodal displacements
   @param nid - node id
   
   20. 8. 2012, JK
*/
void select_noddispl (long lcid,double *r,long nid)
{
  long i,ii;
  
  switch (Mp->tprob){
  case forced_dynamics:
  case mech_timedependent_prob:
  {
    for (i=0;i<Mt->give_ndofn (nid);i++)
    {
      ii=Mt->give_dof(nid,i);
      if (ii<0)   r[i]=Mb->dlc[lcid].get_pd(Mp->time, ii);
      if (ii==0)  r[i]=0.0;
      if (ii>0)   r[i]=Lsrs->lhs[lcid*Ndofm+ii-1];
    }
    break;
  }
  case growing_mech_structure:
    for (i=0;i<Mt->give_ndofn (nid);i++)
    {
      ii=Mt->give_dof(nid,i);
      if (ii<0)   r[i]=Mb->dlc[lcid].get_pd(Mp->time, ii);
      if (ii==0)  r[i]=0.0;
      if (ii>0)   r[i]=Lsrs->lhs[lcid*Ndofm+ii-1];

      if (Mt->nodedispl[nid] && (ii < 1)) 
      // add initial displacements from previous free state of dof 
      // in the case that the dof has got support or prescribed displacement actually
        r[i] += Mt->nodedispl[nid][i];
    }
    break;
  case mat_nonlinear_statics:{
    for (i=0;i<Mt->give_ndofn (nid);i++){
      ii=Mt->give_dof(nid,i);
      if (ii<0)
      {   
        r[i] = 0.0;
        if (Mb->lc[lcid].npd)
          r[i] += Mb->lc[lcid].pd[0-ii-1] * Mp->lambda;
        if (Mb->lc[lcid+1].npd)
          r[i]+= Mb->lc[lcid+1].pd[0-ii-1];
      }
      if (ii==0)  r[i]=0.0;
      if (ii>0)   r[i]=Lsrs->lhs[lcid*Ndofm+ii-1];
    }
    break;
  }
  default:{
    for (i=0;i<Mt->give_ndofn (nid);i++){
      ii=Mt->give_dof(nid,i);
      if (ii<0)   r[i]=Mb->lc[lcid].pd[0-ii-1];
      if (ii==0)  r[i]=0.0;
      if (ii>0)   r[i]=Lsrs->lhs[lcid*Ndofm+ii-1];
    }
  }
  }//  end of the switch (Mp->tprob){
}



/**
   The function selects nodal forces of the nid-th node from the global vector of nodal forces f
   and stores them into %vector nf.
   
   @param lcid - load case id
   @param nf - array of nodal force component
   @param f - load/internal force %vector
   @param nid - node id
   
   4.4. 2016, TKo
*/
void nodforce (double *f, double *nf, long nid)
{
  long i, j, ii, nmn, mnid, ndofn, ndofnm;
  double **ff;
  vector w;
  
  //  the number of DOFs in node
  ndofn = Gtm->give_ndofn(nid);

  if (ndofn<0){
    //  this is a hanging node
    
    //  the number of master nodes
    nmn=0-ndofn;
    ff = new double*[nmn];
    //  loop over the number of master nodes
    for (i=0; i<nmn; i++)
    {
      //  the number of DOFs on master node
      mnid   = Gtm->gnodes[nid].mnodes[i];
      ndofnm = Gtm->give_ndofn(mnid);
      ff[i]  = new double[ndofnm];
      for (j=0; j<Mt->give_ndofn(mnid); j++)
      {
        ii = Mt->give_dof(mnid,j);
        if (ii<=0)   ff[i][j] = Mt->nodes[mnid].r[j];
        else         ff[i][j] = f[ii-1];
      }
    }
    
    //  weights
    reallocv(RSTCKVEC(nmn,w));
    Gtm->approx_weights (Gtm->gnodes[nid].masentity, nid, w, Out);
    
    //  approximation
    //r = new double [ndofnm];
    for (i=0;i<ndofnm;i++){
      nf[i]=0.0;
      for (j=0;j<nmn;j++){
	nf[i]+=ff[j][i]*w[j];
      }
    }
    
    for (i=0;i<0-ndofn;i++){
      delete [] ff[i];
    }
    delete [] ff;
    
    ndofn = ndofnm;
  }else{
    //  this is not a hanging node
    for (i=0; i<Mt->give_ndofn(nid); i++)
    {
      ii = Mt->give_dof(nid,i);
      if (ii<=0)   nf[i]=Mt->nodes[nid].r[i];
      if (ii>0)    nf[i]=f[ii-1];
    }
  }//  end of the statement if (ndofn<0){
}



/**
  The function returns cn-th component of the macro-strain %vector in case of homogeniztaion problem.

  @param lcid - load case id
  @param cn - index of required macro-strain component

  @return The function returns macro-strain components in the argument mstra.

  Created by Tomas Koudelka, 20.1.2015
*/
double macrostrains(long lcid, long cn)
{
  double *d;
  long ncomp = Mm->max_ncompstre;
  double ret = 0.0;
  long j;

  switch (Mp->homog)
  {
    case 3: // macro-stress approach - 
    {       // components of the macro-strains are at last positions in the %vector of unknowns
      d = Lsrs->give_lhs(lcid);
      ret = d[Ndofm-ncomp+cn];
      break;
    }
    case 4: // macro-strain approach - macro-strains are defined in the particular load cases
    {
      if (Mb->dlc)
        ret = Mb->dlc[lcid].get_macrostra(Mp->time, cn);
      else
      {
        for (j=0; j<Mb->nlc; j++)
        {
          if (Mp->tprob == mat_nonlinear_statics)
          {
            ret = Mb->lc[2*lcid].mstrain[cn]*Mp->lambda;
            ret += Mb->lc[2*lcid+1].mstrain[cn];
          }
          else
            ret = Mb->lc[lcid].mstrain[cn];
        }
      }
      break;
    }
    default:
    {
      print_err("unknown type of homogenization is required", __FILE__, __LINE__, __func__);
      abort();
    }
  }
  return ret;
}



/**
  The function returns cn-th component of the macro-stress %vector in case of homogeniztaion problem.

  @param lcid - load case id
  @param cn - index of required macro-strain component

  @return The function returns macro-strain components in the argument mstra.

  Created by Tomas Koudelka, 20.1.2015
*/
double macrostresses(long lcid, long cn)
{
  long ret = 0.0;
  long j;


  switch (Mp->homog)
  {
    case 3: // macro-stress approach - macro-stresses are defined in the particular load cases
    {
      if (Mb->dlc)
        ret = Mb->dlc[lcid].get_macrostre(Mp->time, cn);
      else
      {
        for (j=0; j<Mb->nlc; j++)
        {
          if (Mp->tprob == mat_nonlinear_statics)
          {
            ret = Mb->lc[2*lcid].mstress[cn]*Mp->lambda;
            ret += Mb->lc[2*lcid+1].mstress[cn];
          }
          else
            ret = Mb->lc[lcid].mstress[cn];
        }
      }
      break;
    }
    case 4: // macro-strain approach - 
    {       // components of the macro-stresses are stored in mechmat::mstress
      ret = Mm->mstress[cn];
      break;
    }
    default:
    {
      print_err("unknown type of homogenization is required", __FILE__, __LINE__, __func__);
      abort();
    }
  }
  return ret;
}



/**
  The function extracts nodal force components on one node.

  @param f   - array of components for all nodal forces, it is input parameter
  @param nid - node number (node id)
  @param nf  - vector of nodal force, it is output parameter

  @return The function returns nodal force components at the given node in the %vector nf.

  Created by Tomas Koudelka 1.10.2013
*/
void nodforce (double *f, long nid, vector &nf)
{
  long i,j,ndofn,ndofnm,nmn;
  vector *ff;
  vector w;
  
  //  the number of DOFs in node
  ndofn = Gtm->give_ndofn(nid);
  
  if (ndofn<0)
  { //  this is a hanging node
    
    //  the number of master nodes
    nmn=0-ndofn;
    
    ff = new vector[nmn]; // array of force vectors of the master nodes

    //  loop over the number of master nodes
    for (i=0;i<nmn;i++)
    {
      //  the number of DOFs on master node
      ndofnm = Gtm->give_ndofn(Gtm->gnodes[nid].mnodes[i]);
      reallocv(ndofnm, ff[i]);
      select_nodforce (f, Gtm->gnodes[nid].mnodes[i], ff[i]);
    }
    
    //  weights
    allocv(nmn, w);
    Gtm->approx_weights(Gtm->gnodes[nid].masentity, nid, w, Out);
    
    //  approximation
    for (i=0;i<ndofnm;i++)
    {
      nf[i]=0.0;
      for (j=0; j<nmn; j++)
	nf[i] += ff[j][i]*w[j];
    }
    delete [] ff;
    
    ndofn = ndofnm;
  }
  else
  {
    //  this is not a hanging node
    select_nodforce (f, nid, nf);
  }
}



/**
   The function selects nodal force from the array f and stores its components to 
   the %vector nf.

   @param f   - array of components for all nodal forces, it is input parameter
   @param nid - node id
   @param nf  - %vector of nodal force
   
   1. 10. 2013, TKo
*/
void select_nodforce(double *f, long nid, vector &nf)
{
  long i,ii;
  
  for (i=0;i<Mt->give_ndofn(nid);i++)
  {
    ii=Mt->give_dof(nid,i);
    if (ii < 1)  // there is prescribed DOF
    { 
      if (Mp->reactcomp)    
        nf[i]=Mt->nodes[nid].r[i];
      else                              
        nf[i]=0.0;
    }
    else // there is free DOFs
    {
      if (f != NULL)   
        nf[i]=f[ii-1];
      else                
        nf[i]=0.0;
    }
  }
}



/**
  The function extracts displacements on one node for given problem.
   
  @param lcid - number of load case
  @param idn  - node number
  @param p    - pointer on problem
  @param r    - answer - allocated array for displacement
   
  @return The function returns require vector of nodal displacements in the array r.

  Created JK, 11.10.2003
*/
/*
void noddispl (long lcid, long idn, problem *p, double *r)
{
  long i,ii;
  
  switch (p->mp->tprob)
  {
  case forced_dynamics:
  case growing_mech_structure://tady??!!
  case mech_timedependent_prob:
    {
      for (i=0;i<p->gt->give_ndofn (idn);i++){
        ii = p->gt->give_dof(idn,i);
        if (ii<0)   r[i]=p->mb->dlc[lcid].get_pd(p->mp->time, ii);
        if (ii==0)  r[i]=0.0;
        if (ii>0)   r[i]=p->lsrs->lhs[lcid*p->lsrs->ndof+ii-1];
      }
      break;
    }
    case mat_nonlinear_statics:
    {
      for (i=0;i<p->gt->give_ndofn (idn);i++){
        ii = p->gt->give_dof(idn,i);
        if (ii<0)   r[i]=p->mb->lc[lcid].pd[0-ii-1] * p->mp->lambda + p->mb->lc[lcid+1].pd[0-ii-1];
        if (ii==0)  r[i]=0.0;
        if (ii>0)   r[i]=p->lsrs->lhs[lcid*p->lsrs->ndof+ii-1];
      }
      break;
    }
    default:
    {
      for (i=0;i<p->gt->give_ndofn (idn);i++)
      {
         ii = p->gt->give_dof(idn,i);
        if (ii<0)   r[i]=p->mb->lc[lcid].pd[0-ii-1];
        if (ii==0)  r[i]=0.0;
        if (ii>0)   r[i]=p->lsrs->lhs[lcid*p->lsrs->ndof+ii-1];
      }
    }
  }
}
*/


/**
  The function assembles constraint %matrix of one layer on one node.
   
  @param nid - node id
  @param cid - contact id
  @param lcm - local constraint %matrix (output)
   
  @return The function returns constraint %matrix in the parameter lcm.

  Created by JK, 7.12.2002
*/
void constr_matrix (long nid,long cid,matrix &lcm)
{
  long i,idcs;
  crsectype crst;
  double tl,tu;
  
  i=Gtm->lgnodes[nid].nodes[cid-1];
  crst = Mt->nodes[i].crst;
  idcs = Mt->nodes[i].idcs;
  tl = Mc->give_onethickness (crst,idcs);
  i=Gtm->lgnodes[nid].nodes[cid];
  crst = Mt->nodes[i].crst;
  idcs = Mt->nodes[i].idcs;
  tu = Mc->give_onethickness (crst,idcs);
  
  fillm (0.0,lcm);

  if (Mp->tlm==1){
    lcm[0][0] = -1.0;
    lcm[1][1] = -1.0;
    lcm[2][0] = tl/(-2.0);
    
    lcm[3][0] = 1.0;
    lcm[4][1] = 1.0;
    lcm[5][0] = tu/(-2.0);
  }
  
  if (Mp->tlm==2){
    
    /*
      uplne stara verze
      lcm[0][0] = -1.0;
      lcm[1][1] = -1.0;
      lcm[2][2] = -1.0;
      lcm[1][3] = tl/2.0;
      lcm[0][4] = tl/(-2.0);
      lcm[3][5] = -1.0;
      
      lcm[0][6]  = 1.0;
      lcm[1][7]  = 1.0;
      lcm[2][8]  = 1.0;
      lcm[1][9]  = tu/2.0;
      lcm[0][10] = tu/(-2.0);
      lcm[3][11] = 1.0;
    */
    
    
    lcm[0][0] = -1.0;
    lcm[1][1] = -1.0;
    lcm[2][2] = -1.0;
    lcm[3][1] = tl/2.0;
    lcm[4][0] = tl/(-2.0);
    lcm[5][3] = -1.0;
    
    lcm[6][0]  = 1.0;
    lcm[7][1]  = 1.0;
    lcm[8][2]  = 1.0;
    lcm[9][1]  = tu/2.0;
    lcm[10][0] = tu/(-2.0);
    lcm[11][3] = 1.0;
  }
}



/**
  Function assembles right hand side of system of algebraic equations.
  It is created by prescribed displacements, forces, moments, eigenstrains, etc.
  and it contains nodal forces.
  In case that the flv is not NULL, %vector of load caused by forces only is stored in flv 
   
  Array rhs has to be cleaned in this function, called functions
  do not clean the array rhs.
   
  @param lcid - load case id
  @param rhs - array of right hand side
  @param flv - array of load %vector caused by forces only (default value is NULL)
   
  @return The function returns assembled vectors of righ hand side and load vector caused by force.

  Created by JK,
  Modified by Tomas Koudelka,
*/
void mefel_right_hand_side (long lcid,double *rhs,double *flv)
{

  double *estc; // pointer to righ hand side vector of load case where to add contributions from eigenstrains/eigenstresses
  
  // by default, the eigenstrain/eigenstress contributions are added to the rhs directly
  estc = rhs;

  if (flv)
    nullv(flv, Ndofm);

  switch (Mp->tprob){
  case linear_statics:{
    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs+lcid*Ndofm,Ndofm);
    Mb->lc[lcid].assemble (lcid,rhs+lcid*Ndofm,flv,1.0);
    break;
  }
  case eigen_dynamics:{
    break;
  }
  case forced_dynamics:{
    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs+lcid*Ndofm,Ndofm);
    //Mb->lc[lcid].assemble (lcid,av);
    Mb->dlc[lcid].assemble (lcid,rhs,flv,Ndofm,Mp->time);
    break;
  }
  case linear_stability:{
    break;
  }
  case mat_nonlinear_statics:{
    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs+(lcid*2+0)*Ndofm, Ndofm);
    nullv (rhs+(lcid*2+1)*Ndofm, Ndofm);
	if (flv)
      nullv (flv+(lcid*2+1)*Ndofm, Ndofm);
    Mb->lc[lcid*2+0].assemble (lcid*2+0,rhs+(lcid*2+0)*Ndofm,flv+(lcid*2+0)*Ndofm,1.0);
    Mb->lc[lcid*2+1].assemble (lcid*2+1,rhs+(lcid*2+1)*Ndofm,flv+(lcid*2+1)*Ndofm,1.0);
    estc = rhs+(lcid*2+1)*Ndofm; // eigenstrain/eigenstress contribution will be contained in the constant load vector
    break;
  }
  case geom_nonlinear_statics:{
    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs+(lcid*2+0)*Ndofm, Ndofm);
    nullv (rhs+(lcid*2+1)*Ndofm, Ndofm);
    Mb->lc[lcid*2+0].assemble (lcid*2+0,rhs+(lcid*2+0)*Ndofm,flv,1.0);
    Mb->lc[lcid*2+1].assemble (lcid*2+1,rhs+(lcid*2+1)*Ndofm,flv,1.0);
    estc = rhs+(lcid*2+1)*Ndofm; // eigenstrain/eigenstress contribution will be contained in the constant load vector
    break;
  }
  case mech_timedependent_prob:{
    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs+lcid*Ndofm, Ndofm);
    Mb->dlc[lcid].assemble (lcid,rhs+lcid*Ndofm,flv,Ndofm,Mp->time);
    break;
  }
  case growing_mech_structure:{
    //  lcid must be equal to zero

    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs, Ndofm);
    Mb->dlc[lcid].assemble (0,rhs,flv,Ndofm,Mp->time);
    break;
  }
  case earth_pressure:{
    
    break;
  }
  case layered_linear_statics:{
    
    break;
  }
  case lin_floating_subdomain:{
    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs+lcid*Ndofm,Ndofm);
    Mb->lc[lcid].assemble (lcid,rhs+lcid*Ndofm,flv,1.0);
    break;
  }
  case nonlin_floating_subdomain:{
    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs+(lcid*2+0)*Ndofm, Ndofm);
    nullv (rhs+(lcid*2+1)*Ndofm, Ndofm);
    Mb->lc[lcid*2+0].assemble (lcid*2+0,rhs+(lcid*2+0)*Ndofm,flv,1.0);
    Mb->lc[lcid*2+1].assemble (lcid*2+1,rhs+(lcid*2+1)*Ndofm,flv,1.0);
    estc = rhs+(lcid*2+1)*Ndofm; // eigenstrain/eigenstress contribution will be contained in the constant load vector
    break;
  }
    
  case hemivar_inequal:{
    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs+lcid*Ndofm,Ndofm);
    Mb->lc[lcid].assemble (lcid,rhs+lcid*Ndofm,flv,1.0);
    break;
  }
  case load_balancing:{
    //  components of the vector rhs must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (rhs+lcid*Ndofm,Ndofm);
    Mb->lc[lcid].assemble (lcid,rhs+lcid*Ndofm,flv,1.0);
    break;
  }
    
  default:{
    print_err("unknown problem type is required",__FILE__,__LINE__,__func__);
  }
  }
  
  if (Mp->eigstrains > 0){
    //  contribution to the right hand side (vector of nodal forces)
    //  from prescribed eigenstrains
    
    //  type of eigenstrain is prescribed eigenstrain
    Mm->est=eigstrain;
    
    //  computation of nodal forces caused by eigenstrains
    nodal_eigstrain_forces (lcid,estc,Mp->time);
  }

  if ((Mp->pore_press > 0) && (Mp->pore_press < 3))
  {   
    // For defined values of pore presures or for partially coupled approach,
    // the contribution of nodal forces due to pore pressure is assumed in the right hand side 
    nodal_pore_press_forces (lcid, rhs+lcid*Ndofm, Mp->time);
  }
}



/**
  The function computes all required values
  this function is called e.g. before output print.

  @param lcid - load case id

  @return The function does not return anything.

  Created by JK, 1.4.2004
  Modified by TKo, 07.2008
*/
void compute_req_val (long lcid)
{
  long aux, i, j;

  // strain computaion
  if (Mp->straincomp == 1)
  {
    if (Mp->strainstate==0)
    {
      compute_ipstrains (lcid);
      //  indicator of strain computation
      //  it means, strains at integration points have been computed
      Mp->strainstate=1;
    }
    if (Mp->strainpos == 2)
      compute_nodestrains (lcid);
  }
  
  // stress computaion
  if (Mp->stresscomp == 1)
  {
    // strains have to be computed if they are not computed yet
    if (Mp->strainstate==0)
    {
      compute_ipstrains (lcid);
      //  indicator of strain computation
      //  it means, strains at integration points have been computed
      Mp->strainstate=1;
    }
    if (Mp->stressstate==0)
    {
      compute_ipstresses (lcid);
      //  indicator of stress computation
      //  it means, stresses at integration points have not been computed
      Mp->stressstate=1;
      //  indicator of computation of other array
      //  it means, stresses at integration points have not been computed
      Mp->otherstate=1;
    }
    if (Mp->stresspos == 2)
      compute_nodestresses (lcid);
    if (Mp->homog==4) // macro-strain approach
    {
      memset(Mm->mstress, 0, sizeof(*Mm->mstress)*Mm->max_ncompstre);

      for (i=0; i<Mm->tnip; i++)
      {
        for (j=0; j<Mm->max_ncompstre; j++)
          Mm->mstress[i]+=Mm->ip[i].stress[j]*Mm->ipv[i];
      }
      for (j=0; j<Mm->max_ncompstre; j++)
        Mm->mstress[i]/=Mt->domvol;
    }
  }
  
  // other values computation
  if ((Mp->othercomp == 1) && (Mp->otherpos == 2) && (Mp->otherstate==1))
    compute_nodeothers ();
  
  //  reactions computation
  if (Mp->reactcomp==1){
    // strains have to be computed at integration points
    if (Mp->strainstate==0)
    {
      compute_ipstrains (lcid);
      //  indicator of strain computation
      //  it means, strains at integration points have been computed
      Mp->strainstate=1;
    }
    if (Mp->stressstate==0)
    {
      compute_ipstresses (lcid);
      //  indicator of stress computation
      //  it means, stresses at integration points have not been computed
      Mp->stressstate=1;
      //  indicator of computation of other array
      //  it means, stresses at integration points have not been computed
      Mp->otherstate=1;
    }
    aux = Mp->strcomp;
    Mp->strcomp = 0; // stresses have been already computed, no stress computation 
                     // is required in the integration of forces on the elements
    Mt->null_react();
    switch (Mp->tprob){
    case linear_statics:
    case load_balancing:
    case lin_floating_subdomain:
      Mb->lc[lcid].compute_reactions(lcid, 1.0, yes);
      break;
    case mat_nonlinear_statics:
    case earth_pressure:
    case nonlin_floating_subdomain:
      // Reactions from proportional load case involve contributions from internal forces contributions
      Mb->lc[lcid].compute_reactions(lcid, Mp->lambda, yes);
      // Reaction contributions from constant load case must NOT involve contributions from internal forces contributions
      // otherwise they will be assumed twice. Parameter lcid is not used in the such case
      Mb->lc[lcid+1].compute_reactions(lcid, 1.0, no); 
      break;
    case forced_dynamics:
    case mech_timedependent_prob:
    case growing_mech_structure:
      Mb->dlc[lcid].compute_reactions(lcid);
      break;
    case eigen_dynamics:
      break; 
    default:
      print_err("unknown type of problem is required",__FILE__,__LINE__,__func__);
    }
    Mp->strcomp = aux; // restore the default state
  }

  // computation of strains have to be last due to total strains
  // all previous computations need strains - eigenstrains
  if (Mp->straincomp == 1){
    if (Mp->strainstate==0)
    {
      compute_ipstrains (lcid);
      //  indicator of strain computation
      //  it means, strains at integration points have been computed
      Mp->strainstate=1;
    }
    // in previous statements, strains were modified
    // such that the eigstrains and temprstrains were subtracted
    // total strains have to be computed again    
    Mm->totstrains ();
    if (Mp->strainpos == 2)
      compute_nodestrains (lcid);
  }
}



/**
  The function transforms nodal displacements form the nodal local coordinate system
  to the global one.

  @param lcid - load case id

  @return The function does not return anything.

  Created by JK,
*/
void local_global_displ_transf (long lcid)
{
  long i,j,k,ndofn;
  long *cn;
  vector l,g;
  matrix tm(3,3);
  
  for (i=0;i<Mt->nn;i++){
    if (Mt->nodes[i].transf>0){

      //fprintf (Outm,"\n uzel %ld   %ld",i,Mt->nodes[i].transf);

      ndofn = Mt->give_ndofn (i);
      allocv (ndofn,l);
      allocv (ndofn,g);
      
      noddispl (lcid,l.a,i);

      tm[0][0]=Mt->nodes[i].e1[0];
      tm[1][0]=Mt->nodes[i].e1[1];
      tm[2][0]=Mt->nodes[i].e1[2];
      
      tm[0][1]=Mt->nodes[i].e2[0];
      tm[1][1]=Mt->nodes[i].e2[1];
      tm[2][1]=Mt->nodes[i].e2[2];
      
      tm[0][2]=Mt->nodes[i].e3[0];
      tm[1][2]=Mt->nodes[i].e3[1];
      tm[2][2]=Mt->nodes[i].e3[2];
      
      mxv(tm,l,g);
      
      cn = new long [ndofn];
      Mt->give_node_code_numbers (i,cn);
      
      for (j=0;j<ndofn;j++){
	k=cn[j]-1;
	if (k>-1)
	  Lsrs->lhs[k]=g[j];
      }
      delete [] cn;
      destrv (l);
      destrv (g);
    }
  }
}



/**
  The function compute stresses due to initial displacements on the interfaces 
  between new and old part of the structure. It is used due to correct graphics 
  representation of the deformed shape in case of the growing structures.

  Parameters:
  @param lcid - load case id

  @return The function deos not return anything but changes stress array of integration 
          points on active elements.

  Created by TKo, 7.6.2013
*/
void stress_initdispl(long lcid)
{
  long i, j, ne, tnip, ncomp, ipp;
  matrix d;
  vector eps, sig;

  compute_ipstrains(lcid);
  ne = Mt->ne;
  for (i=0;i<ne;i++)
  {
    if (Gtm->leso[i]==1)
    {
      tnip = Mt->give_tnip(i);
      ipp = Mt->elements[i].ipp[0][0];
      for (j=0; j<tnip; j++)
      {
        ncomp = Mm->ip[ipp].ncompstr;
        reallocm(ncomp, ncomp, d);
        reallocv(ncomp, eps);
        reallocv(ncomp, sig);     
        Mm->matstiff(d, ipp);
        Mm->givestrain(lcid, ipp, eps);
        mxv(d, eps, sig);
        Mm->storestress(lcid, ipp, sig);
        ipp++;
      }
    }
  }

  //  indicator of strain computation
  //  it means, strains at integration points have been computed
  Mp->strainstate=1;
  //  indicator of stress computation
  //  it means, stresses at integration points have been computed
  Mp->stressstate=1;
  //  indicator of computation of other array
  //  it means, stresses at integration points have been computed
  Mp->otherstate=1;
}

